///////////////////////////////////////////////////////////////////////////////
// This source file is part of the LuaPlus source distribution and is Copyright
// 2002 by Joshua C. Jensen (jjensen@workspacewhiz.com).
//
// The latest version may be obtained from http://wwhiz.com/LuaPlus/.
//
// The code presented in this file may be freely used and modified for all
// non-commercial and commercial purposes.
///////////////////////////////////////////////////////////////////////////////
#define BUILDING_LUAPLUS
#include "LuaLink.h"
LUA_EXTERN_C_BEGIN
#include "lobject.h"
LUA_EXTERN_C_END
#include "LuaPlus.h"
#include <stdio.h>
#include <string.h>
#ifdef WIN32
#include <windows.h>
#undef GetObject
#endif WIN32
#pragma warning(disable: 4786)
#pragma warning(disable: 4702)
#include <list>
#include <string>
#include <algorithm>
#include "LuaPlusFunctions.h"
#include "LuaPlusInternals.h"
#pragma warning(disable: 4100)

//-----------------------------------------------------------------------------
LUA_EXTERN_C_BEGIN
#include "lualib.h"
#include "lfunc.h"
#include "lgc.h"
#include "lstate.h"
#include "lvm.h"
#include "include/Lua.h"
#include "include/lauxlib.h"
LUA_EXTERN_C_END


NAMESPACE_LUA_BEGIN
LUA_EXTERN_C int str_format_helper (luaL_Buffer* b, lua_State *L, int arg);
NAMESPACE_LUA_END


namespace LuaPlus
{

USING_NAMESPACE_LUA

static void LSLock(void* data);
static void LSUnlock(void* data);

void LuaPlusGCFunction(GCState*);

#ifndef api_check
#define api_check(L, o)		(assert(o))
#endif

#define api_checknelems(L, n)	api_check(L, (n) <= (L->top - L->ci->base))

//#define api_incr_top(L)   (api_check(L, L->top<L->ci->top), L->top++)

#define api_incr_top(L) \
	{if (L->top >= L->ci->top) lua_checkstack(L, 1); L->top++;}

LUA_API void lua_pushcfunctorclosure (lua_State *L, LSFunctor fn, int n)
{
	Closure *cl;
	lua_lock(L);
	api_checknelems(L, n);
	cl = luaF_newCclosure(L, n);
	cl->c.f = NULL;
	cl->c.functiontype = 1;
	CClosureEx* closureEx = (CClosureEx*)(&cl->c + 1);
	::new(&closureEx->f1) LSFunctor;
	LSFunctor* functor = (LSFunctor*)&closureEx->f1;
	*functor = fn;
	L->top -= n;
	while (n--)
		setobj(getcclosureupvalue(L, cl->c, n), L->top+n);
	setclvalue(L->top, cl);
	api_incr_top(L);
	lua_unlock(L);
}


LUA_API void lua_pushcclosureex (lua_State *L, LuaStateCFunction fn, int n) {
  Closure *cl;
  lua_lock(L);
  api_checknelems(L, n);
  cl = luaF_newCclosure(L, n);
  cl->c.functiontype = 2;
  CClosureEx* closureEx = (CClosureEx*)(&cl->c + 1);
  closureEx->f2 = fn;
  L->top -= n;
  while (n--)
    setobj(getcclosureupvalue(L, cl->c, n), L->top+n);
  setclvalue(L->top, cl);
  api_incr_top(L);
  lua_unlock(L);
}


static int LuaState_CClosureManager(lua_State *L, int closureState, CClosure* closure)
{
	CClosureEx* closureEx = (CClosureEx*)(closure + 1);
	if ( closureState == CCLOSUREMANAGER_CALL_CLOSURE )
	{
		int n = 0;
		LuaState* state = (LuaState*)lua_getstateuserdata( L );

		if ( closure->functiontype == 1 )
		{
			LSFunctor* functor = (LSFunctor*)&closureEx->f1;
			n = (*functor)( state, (LuaStackObject*)&state->m_callbackArgs );
		}
		else if ( closure->functiontype == 2 )
		{
			n = (*closureEx->f2)( state, (LuaStackObject*)&state->m_callbackArgs );
		}

		return n;
	}
	else if ( closureState == CCLOSUREMANAGER_INIT_CLOSURE )
	{
		memset( closureEx, 0, sizeof( CClosureEx ) );
	}
	else if ( closureState == CCLOSUREMANAGER_FREE_CLOSURE )
	{
	}

	return 0;
}


LUA_API void lua_pushtobject(lua_State *L, void* tobject)
{
	TObject* tobj = (TObject*)tobject;
	lua_lock(L);
	*L->top = *tobj;
	api_incr_top(L);
	lua_unlock(L);
}


/*static*/ LuaState* LuaState::Create( bool initStandardLibrary, bool multithreaded )
{
	lua_ReallocFunction reallocFunc;
	lua_FreeFunction freeFunc;
	void* data;
	lua_getdefaultmemoryfunctions(&reallocFunc, &freeFunc, &data);
	LuaState* state = (LuaState*)(*reallocFunc)(NULL, 0, sizeof(LuaState), data, "LuaState", 0);
	::new(state) LuaState(initStandardLibrary, multithreaded);
	return state;
}


/*static*/ LuaState* LuaState::CastState( lua_State* L )
{
	LuaState* state = (LuaState*)lua_getstateuserdata( L );
	return state;
}


/*static*/ void LuaState::Destroy( LuaState* state )
{
	lua_FreeFunction freeFunc = state->m_state->l_G->freeFunc;
	void* data = state->m_state->l_G->memData;
	state->~LuaState();
	(*freeFunc)(state, sizeof(LuaState), data);
}


LuaState* LuaState::CreateThread(LuaState* parentState)
{
	lua_ReallocFunction reallocFunc;
	lua_FreeFunction freeFunc;
	void* data;
	lua_getdefaultmemoryfunctions(&reallocFunc, &freeFunc, &data);
	LuaState* state = (LuaState*)(*reallocFunc)(NULL, 0, sizeof(LuaState), data, "LuaState", 0);
	::new(state) LuaState(parentState, false);
	return state;
}

	
/**
**/
LuaState::LuaState(bool initStandardLibrary, bool multithreaded)
{
	m_state = lua_open();
	m_ownState = true;

	lua_setcclosureinfo( m_state, sizeof( CClosureEx ), LuaState_CClosureManager );
	lua_setusergcfunction(m_state, LuaPlusGCFunction);

	SetupStateEx();

	GetHeadObject()->m_prev = NULL;
	GetHeadObject()->m_next = GetTailObject();
	GetTailObject()->m_prev = GetHeadObject();
	GetTailObject()->m_next = NULL;

	if (multithreaded)
	{
#ifdef WIN32
		// What about clean up?
		CRITICAL_SECTION* cs = new CRITICAL_SECTION;
		::InitializeCriticalSection(cs);
		lua_setlockfunctions(m_state, LSLock, LSUnlock, cs);
#endif WIN32
	}
	
	Init(initStandardLibrary);
}


/**
**/
LuaState::LuaState(LuaState* parentState, bool initStandardLibrary)
{
	m_state = lua_newthread(*parentState);
	m_ownState = false;			// Threads are not closeable.

	GetHeadObject()->m_prev = NULL;
	GetHeadObject()->m_next = GetTailObject();
	GetTailObject()->m_prev = GetHeadObject();
	GetTailObject()->m_next = NULL;

	m_threadObj = LuaStackObject(parentState, parentState->GetTop());
	parentState->Pop();

	SetupStateEx();
}


/**
**/
LuaState::LuaState( lua_State* state )
{
	m_state = state;

	m_ownState = false;

	SetupStateEx();
}


/**
	The LuaState class assumes ownership of the lua_State pointer and
	destroys it when destroyed.
**/
LuaState::~LuaState()
{
	if ( m_state  &&  m_ownState )
	{
		lua_close( m_state );
	}
}


/**
**/
void LuaState::SetupStateEx()
{
	for ( int i = 0; i < LUA_MINSTACK; ++i )
	{
		m_callbackArgs[ i ].m_state = this;
		m_callbackArgs[ i ].m_stackIndex = i;
	}

	lua_setstateuserdata( m_state, this );

	m_tweakMeToSeeStuff.m_state = this;
	m_tweakMeToSeeStuff.m_stackIndex = 1;
}

static int LS_LOG( LuaState* state, LuaStackObject* args )
{
	lua_State* L = *state;
	int n = lua_gettop(L);  /* number of arguments */
	int i;
	lua_getglobal(L, "towstring");
	lua_getglobal(L, "tostring");
	for (i=1; i<=n; i++) {
		const char *s = NULL;
		const wchar_t *ws = NULL;
		if (lua_type(L, i) != LUA_TWSTRING)
		{
			lua_pushvalue(L, -1);  /* function to be called */
			lua_pushvalue(L, i);   /* value to print */
			lua_call(L, 1, 1);
			s = lua_tostring(L, -1);  /* get result */
			if (s == NULL)
				return luaL_error(L, "`tostring' must return a string to `print'");
		}
		else
		{
			lua_pushvalue(L, -2);  /* function to be called */
			lua_pushvalue(L, i);   /* value to print */
			lua_call(L, 1, 1);
			ws = lua_towstring(L, -1);  /* get result */
			if (ws == NULL)
				return luaL_error(L, "`tostring' must return a string to `print'");
		}
		if (i>1)
		{
#ifdef WIN32
			OutputDebugStringA("\t");
#else
			fputs("\t", stdout);
#endif
		}
		if (s)
		{
#ifdef WIN32
			OutputDebugStringA(s);
#else
			fputs(s, stdout);
#endif
		}
		else
		{
#ifdef WIN32
			OutputDebugStringW(ws);
#else
			fputws(ws, stdout);
#endif
		}
		lua_pop(L, 1);  /* pop result */
	}

#ifdef WIN32
	OutputDebugStringA("\n");
#else
	fputs("\n", stdout);
#endif

	return 0;
}


class LUA_API LuaStateOutputDebugStringFile : public LuaStateOutFile
{
public:
	LuaStateOutputDebugStringFile() {}
	virtual ~LuaStateOutputDebugStringFile() {}

	virtual bool Open(const char* fileName)
	{
		fileName;
		return true;
	}

	virtual void Close()
	{
	}

	virtual void Print(const char* str, ...)
	{
		char message[800];
		va_list arglist;

		va_start(arglist, str);
		vsprintf(message, str, arglist);
		va_end(arglist);

#ifdef WIN32		
		OutputDebugString(message);
#else !WIN32
		puts(message);
#endif WIN32
	}

protected:
};


/**
**/
static int FatalError( lua_State* state )
{
	const char* err = lua_tostring(state, 1);
#ifdef WIN32
	OutputDebugString(err);
#else !WIN32
	puts(err);
#endif WIN32

	throw -1;
}


/**
**/
static void LSLock(void* data)
{
#ifdef WIN32
	CRITICAL_SECTION* cs = (CRITICAL_SECTION*)data;
	::EnterCriticalSection(cs);
#endif WIN32
}


/**
**/
static void LSUnlock(void* data)
{
#ifdef WIN32
	CRITICAL_SECTION* cs = (CRITICAL_SECTION*)data;
	::LeaveCriticalSection(cs);
#endif WIN32
}


/**
**/
static int LS_LuaDumpObject( LuaState* state, LuaStackObject* args )
{
	LuaStateOutFile* file = NULL;
	LuaStateOutFile regFile;
	LuaStateOutputDebugStringFile odsFile;
    LuaStateOutString sFile;

	LuaStackObject& fileObj = args[1];
    if ( fileObj.IsNil() ) {
        file = &sFile;
    }
	else if ( fileObj.IsUserData() )
	{
		FILE* stdioFile = (FILE *)fileObj.GetUserData();
		file = &regFile;
		file->Assign( stdioFile );
	}
	else if ( fileObj.IsString() )
	{
		const char* fileName = fileObj.GetString();
		if ( strcmp( fileObj.GetString(), "@" ) == 0 )
		{
			file = &odsFile;
		}
		else
		{
			file = &regFile;
			if (!file->Open(fileName))
			{
				state->PushBoolean(false);
				return 1;
			}
		}
	}

	LuaStackObject& nameObj = args[2];
	LuaObject valueObj = args[3];
	LuaStackObject& alphabeticalObj = args[4];
	LuaStackObject& indentLevelObj = args[5];
	LuaStackObject& maxIndentLevelObj = args[6];
	LuaStackObject& writeAllObj = args[7];
	bool writeAll = writeAllObj.IsBoolean() ? writeAllObj.GetBoolean() : false;
	bool alphabetical = alphabeticalObj.IsBoolean() ? alphabeticalObj.GetBoolean() : false;
	int maxIndentLevel = maxIndentLevelObj.IsNumber() ? maxIndentLevelObj.GetInteger() : 0xFFFFFFFF;

	state->DumpObject(*file, nameObj.GetString(), valueObj, alphabetical, indentLevelObj.GetInteger(), maxIndentLevel, writeAll, false);
    
    if (fileObj.IsNil()) { //sFile
        state->PushString(sFile.GetBuff());
        return 1;
    } else 
        return 0;
}


/**
**/
static int LS_LuaDumpFile( LuaState* state, LuaStackObject* args )
{
	return LS_LuaDumpObject(state, args);
}


/**
**/
void LuaState::Init( bool initStandardLibrary )
{
	// Register some basic functions with Lua.
	if (initStandardLibrary)
	{
		// A bug in the Lua 5.0 beta causes stack entries to be left behind.
		LuaAutoBlock autoBlock(this);
		lua_baselibopen(m_state);
		lua_tablibopen( m_state );
		lua_iolibopen(m_state);
		lua_strlibopen(m_state);
		lua_wstrlibopen(m_state);
		lua_mathlibopen(m_state);
		lua_dblibopen(m_state);

		ScriptFunctionsRegister( this );

		GetGlobals().Register("LuaDumpObject", LS_LuaDumpObject);
		GetGlobals().Register("LuaDumpFile", LS_LuaDumpFile);
	}

	GetGlobals().Register("LOG", LS_LOG);
	GetGlobals().Register("_ALERT", LS_LOG);

	lua_atpanic( m_state, FatalError );
}


LuaObject LuaState::GetGlobals()
{
	return LuaObject( this, gt(m_state) );
}

	
static void luaI_addquotedbinary (LuaStateOutFile& file, const char* s, int l)
{
	file.Print("%c", '"');
	while (l--)
	{
		switch (*s)
		{
			case '"':  case '\\':
				file.Print("\\%c", *s);
				break;
			case '\a':		file.Print("\\a");		break;
			case '\b':		file.Print("\\b");		break;
			case '\f':		file.Print("\\f");		break;
			case '\n':		file.Print("\\n");		break;
			case '\r':		file.Print("\\r");		break;
			case '\t':		file.Print("\\t");		break;
			case '\v':		file.Print("\\v");		break;
			default:
				if (isprint((unsigned char)*s))
					file.Print("%c", *s);
				else
				{
					file.Print("\\x%02x", (unsigned long)(unsigned char)*s);
				}
		}
		s++;
	}
	file.Print("%c", '"');
}


static void luaI_addquotedwidebinary (LuaStateOutFile& file, const wchar_t* s, int l)
{
	file.Print("L\"");
	while (l--)
	{
		switch (*s)
		{
			case '"':
			case '\\':
				file.Print("\\%c", *s);
				break;
			case '\a':		file.Print("\\a");		break;
			case '\b':		file.Print("\\b");		break;
			case '\f':		file.Print("\\f");		break;
			case '\n':		file.Print("\\n");		break;
			case '\r':		file.Print("\\r");		break;
			case '\t':		file.Print("\\t");		break;
			case '\v':		file.Print("\\v");		break;
			default:
				if (*s < 256  &&  isprint((unsigned char)*s))
				{
					file.Print("%c", *s);
				}
				else
				{
					file.Print("\\x%04x", (unsigned long)*s);
				}
		}
		s++;
	}
	file.Print("\"");
}


#define bufflen(B)	((B)->p - (B)->buffer)
	
static int LS_LuaFilePrint( LuaState* state, LuaStackObject* args )
{
	LuaStateOutFile* file = (LuaStateOutFile*)args[ 1 ].GetUserData();
	
	luaL_Buffer b;
	str_format_helper(&b, *state, 2);

	size_t l = bufflen(&b);
	if (l != 0)
	{
		if (b.isWide)
			assert(0);
		else
		{
			luaL_putchar(&b, 0);
			file->Print(b.buffer);
		}
	}

	return 0;
}


static int LS_LuaFileIndent( LuaState* state, LuaStackObject* args )
{
	LuaStateOutFile* file = (LuaStateOutFile*)args[ 1 ].GetUserData();
	int indentLevel = (int)args[ 2 ].GetNumber();
	file->Indent(indentLevel);

	return 0;
}


bool LuaState::CallFormatting(LuaObject& tableObj, LuaStateOutFile& file, int indentLevel,
		bool writeAll, bool alphabetical, bool writeTablePointers, unsigned int maxIndentLevel)
{
	LuaObject metaTableObj = tableObj.GetMetaTable();
	if (metaTableObj.IsNil())
		return false;

	LuaObject formattedWriteObj = metaTableObj["FormattedWrite"];
	if (!formattedWriteObj.IsFunction())
		return false;

	LuaState* state = tableObj;
	
	{
		LuaObject funcObj = state->GetGlobals()["LuaFilePrint"];
		if (funcObj.IsNil())
		{
			state->GetGlobals().Register("LuaFilePrint", LS_LuaFilePrint);
		}

		funcObj = state->GetGlobals()["LuaFileIndent"];
		if (funcObj.IsNil())
		{
			state->GetGlobals().Register("LuaFileIndent", LS_LuaFileIndent);
		}
	}

	formattedWriteObj.PCall("uobiibb", &file, &tableObj, alphabetical, indentLevel, maxIndentLevel, writeAll, writeTablePointers);

	return true;
}


/**
	Writes a Lua object to a text file.
**/
bool LuaState::DumpObject(LuaStateOutFile& file, const char* name, LuaObject& value,
						 bool alphabetical, int indentLevel, unsigned int maxIndentLevel,
						 bool writeAll, bool writeTablePointers)
{
	// If there is nothing in the variable, then don't write it.
	if (value.IsNil())
		return false;

	// If the variable is user data or a function, then don't write it.
	if (value.IsUserData()  ||  value.IsFunction()  ||  value.IsCFunction())
	{
		if (writeAll  &&  name)
		{
			if (value.IsUserData())
				file.Print("-- %s = '!!!USERDATA!!!'\n", name);
			else if (value.IsFunction())
			{
				lua_Debug ar;
				value.PushStack();
				lua_getinfo(m_state, ">S", &ar);
//				printf("%d\n", ar.linedefined);
				file.Print("-- %s = '!!!FUNCTION!!! %s %d'\n", name, ar.source, ar.linedefined);
			}
			else
				file.Print("-- %s = '!!!CFUNCTION!!!'\n", name);
			return true;
		}
		return false;
	}

	// Indent the line the number of spaces for the current indentation level.
	const unsigned int INDENT_SIZE = 1;
	const unsigned int indentSpaces = (indentLevel == -1 ? 0 : indentLevel) * INDENT_SIZE;
	if ((unsigned int)indentLevel < maxIndentLevel)
		file.Indent(indentSpaces);
	else
		file.Print(" ");
	
	// If the object has a name, write it out.
	if (name)
		file.Print("%s = ", name);

	// If the object's value is a number, write it as a number.
	if (value.IsBoolean())
		file.Print("%s", value.GetBoolean() ? "true" : "false");

	else if (value.IsNumber())
		file.Print("%.16g", value.GetNumber());

	// Or if the object's value is a string, write it as a quoted string.
	else if (value.IsString())
	{
		luaI_addquotedbinary(file, value.GetString(), value.StrLen());
	}

	// Or if the object's value is a string, write it as a quoted string.
	else if (value.IsWString())
	{
		luaI_addquotedwidebinary(file, value.GetWString(), value.StrLen());
	}

	// Otherwise, see if the object's value is a table.
	else if (value.IsTable())
	{
		bool calledFormatting =
				CallFormatting(value, file, indentLevel, writeAll, alphabetical, writeTablePointers, maxIndentLevel);
		if (!calledFormatting)
		{
			// Write the table header.
			if (indentLevel != -1)
			{
				if ((unsigned int)indentLevel + 1 < maxIndentLevel)
				{
					file.Print("\n");
					file.Indent(indentSpaces);
				}
				if (writeTablePointers)
					file.Print("{ --%8x\n", value.GetLuaPointer());
				else
					file.Print("{");
				if ((unsigned int)indentLevel + 1 < maxIndentLevel)
				{
					file.Print("\n");
				}
			}

			// Rename, just for ease of reading.
			LuaObject table = value;

			// upperIndex is the upper index value of a sequential numerical array
			// items.
			int upperIndex = 1;
			bool wroteSemi = false;
			bool hasSequential = false;

			// Block to search for array items.
			{
				// Grab index 1 and index 2 of the table.
				LuaObject value1 = table[1];
				LuaObject value2 = table[2];

				// If they both exist, then there is a sequential list.
				if (!value1.IsNil()  &&  !value2.IsNil())
				{
					// Cycle through the list.
					bool headSequential = true;
					for (; ; ++upperIndex)
					{
						// Try retrieving the table entry at upperIndex.
						LuaObject value = table[upperIndex];

						// If it doesn't exist, then exit the loop.
						if (value.IsNil())
							break;

						// Only add the comma and return if not on the head item.
						if (!headSequential  &&  indentLevel != -1)
						{
							file.Print(",");
							if ((unsigned int)indentLevel + 1 < maxIndentLevel)
							{
								file.Print("\n");
							}
						}
						
						// Write the object as an unnamed entry.
						DumpObject(file, NULL, value, alphabetical, indentLevel + 1, maxIndentLevel, writeAll, writeTablePointers);

						// We've definitely passed the head item now.
						headSequential = false;
					}
				}
			}

			// Did we find any sequential table values?
			if (upperIndex > 1)
			{
				hasSequential = true;
			}
			
			if (alphabetical)
			{
				std::list<std::string> keyNames;

				// Cycle through the table.
				for (LuaTableIterator it(table); it; ++it)
				{
					char keyName[255];

					// Retrieve the table entry's key and value.
					LuaObject key = it.GetKey();

					// Is the key a number?
					if (key.IsNumber())
					{
						// Yes, were there sequential array items in this table?
						if (hasSequential)
						{
							// Is the array item's key an integer?
							float realNum = key.GetNumber();
							int intNum = (int)realNum;
							if (realNum == (float)intNum)
							{
								// Yes.  Is it between 1 and upperIndex?
								if (intNum >= 1  &&  intNum < upperIndex)
								{
									// We already wrote it as part of the sequential
									// list.
									continue;
								}
							}
						}

						// Build the table entry name for the number.
						sprintf(keyName, "[%.16g]", key.GetNumber());
					}
					else
					{
						// Build the table entry name for the string key name.
						const char* ptr = key.GetString();
						bool isAlphaNumeric = true;
						if (isdigit(*ptr))
							isAlphaNumeric = false;
						while (*ptr)
						{
							if (!isalnum(*ptr)  &&  *ptr != '_')
							{
								isAlphaNumeric = false;
								break;
							}
							ptr++;
						}
						if (isAlphaNumeric)
							strcpy(keyName, key.GetString());
						else
							sprintf(keyName, "[\"%s\"]", key.GetString());
					}

					keyNames.push_back(keyName);
				}

				keyNames.sort();

				if (keyNames.size() > 0)
				{
					// If we wrote a sequential list, the value we're about to write
					// is not nil, and we haven't written the semicolon to separate
					// the sequential table entries from the keyed table entries...
					if (hasSequential  &&  indentLevel != -1)
					{
						// Then add a comma (for good measure) and the semicolon.
						file.Print(", ");
						if ((unsigned int)indentLevel + 1 < maxIndentLevel)
						{
							file.Print("\n");
						}
						wroteSemi = true;
					}
				}

				for(std::list<std::string>::iterator it = keyNames.begin(); it != keyNames.end(); ++it)
				{
					const std::string& keyFullName = (*it);

					LuaObject value = table;		// Temp
					if (keyFullName[0] == '[')
					{
						if (keyFullName[1] == '"')
						{
							std::string keyName = keyFullName.substr(2, keyFullName.size() - 4);
							value = table[keyName.c_str()];
						}
						else
						{
							std::string keyName = keyFullName.substr(1, keyFullName.size() - 2);
							int number = atoi(keyName.c_str());
							value = table[number];
						}
					}
					else
					{
						value = table[keyFullName.c_str()];
					}

					// Write the table entry.
					bool ret = DumpObject(file, keyFullName.c_str(), value, alphabetical, indentLevel + 1, maxIndentLevel, writeAll, writeTablePointers);

					// Add a comma after the table entry.
					if (indentLevel != -1  &&  ret)
					{
						file.Print(",");
						if ((unsigned int)indentLevel + 1 < maxIndentLevel)
						{
							file.Print("\n");
						}
					}
				}
			}
			else
			{
				// Cycle through the table.
				for (LuaTableIterator it(table); it; ++it)
				{
					char keyName[255];

					// Retrieve the table entry's key and value.
					LuaObject key = it.GetKey();

					// Is the key a number?
					if (key.IsNumber())
					{
						// Yes, were there sequential array items in this table?
						if (hasSequential)
						{
							// Is the array item's key an integer?
							float realNum = key.GetNumber();
							int intNum = (int)realNum;
							if (realNum == (float)intNum)
							{
								// Yes.  Is it between 1 and upperIndex?
								if (intNum >= 1  &&  intNum < upperIndex)
								{
									// We already wrote it as part of the sequential
									// list.
									continue;
								}
							}
						}

						// Build the table entry name for the number.
						sprintf(keyName, "[%.16g]", key.GetNumber());
					}
					else
					{
						// Build the table entry name for the string key name.
						const char* ptr = key.GetString();
						bool isAlphaNumeric = true;
						if (isdigit(*ptr))
							isAlphaNumeric = false;
						while (*ptr)
						{
							if (!isalnum(*ptr)  &&  *ptr != '_')
							{
								isAlphaNumeric = false;
								break;
							}
							ptr++;
						}
						if (isAlphaNumeric)
							strcpy(keyName, key.GetString());
						else
							sprintf(keyName, "[\"%s\"]", key.GetString());
					}

					// If we wrote a sequential list, the value we're about to write
					// is not nil, and we haven't written the semicolon to separate
					// the sequential table entries from the keyed table entries...
					if (hasSequential  &&  !value.IsNil()  &&  !wroteSemi)
					{
						// Then add a comma (for good measure) and the semicolon.
						if (indentLevel != -1)
						{
							file.Print(", ");
							if ((unsigned int)indentLevel + 1 < maxIndentLevel)
							{
								file.Print("\n");
							}
						}
						wroteSemi = true;
					}

					// Write the table entry.
					bool ret = DumpObject(file, keyName, it.GetValue(), alphabetical, indentLevel + 1, maxIndentLevel, writeAll, writeTablePointers);

					// Add a comma after the table entry.
					if (ret  &&  indentLevel != -1)
					{
						file.Print(",");
						if ((unsigned int)indentLevel + 1 < maxIndentLevel)
						{
							file.Print("\n");
						}
					}
				}
			}

			// If we wrote a sequential list and haven't written a semicolon, then
			// there were no keyed table entries.  Just write the final comma.
			if (hasSequential  &&  !wroteSemi  &&  indentLevel != -1)
			{
				file.Print(",");
				if ((unsigned int)indentLevel + 1 < maxIndentLevel)
				{
					file.Print("\n");
				}
			}
			
			// Indent, with the intent of closing up the table.
			file.Indent(indentSpaces);

			// If the indentation level is 0, then we're at the root position.
			if (indentLevel == 0)
			{
				// Add a couple extra returns for readability's sake.
				file.Print("}");
				if ((unsigned int)indentLevel + 1 < maxIndentLevel)
				{
					file.Print("\n\n", file);
				}
			}
			else if (indentLevel > 0)
			{
				// Close the table.  The comma is written when WriteObject()
				// returns from the recursive call.
				file.Print("}");
			}
		}
	}

	// If the indentation level is at the root, then add a return to separate
	// the lines.
	if (indentLevel == 0)
	{
		if ((unsigned int)indentLevel < maxIndentLevel)
		{
			file.Print("\n");
		}
	}

	return true;
}


/**
	Save the complete script state.
**/
bool LuaState::DumpFile(const char* filename, const char* name, LuaObject& value,
						 bool alphabetical, int indentLevel, unsigned int maxIndentLevel,
						 bool writeAll, bool writeTablePointers)
{
	// Open the text file to write the script state to.
	LuaStateOutFile file;
	if (!file.Open(filename))
		return false;

	// Run through all the globals.
	DumpObject(file, name, value, alphabetical, indentLevel, maxIndentLevel,
			writeAll, writeTablePointers);

	return true;
}


/**
	Adds [indentLevel] number of spaces to the file.
**/
void LuaStateOutFile::Indent( unsigned int indentLevel )
{
	// Write out indentation.
	char spaces[500];
	for (unsigned int i = 0; i < indentLevel; ++i)
		spaces[i] = '\t';
	spaces[i] = 0;
	Print(spaces);
}


///////////////////////////////////////////////////////////////////////////////
LuaStackObject LuaStackObject::PCall(int nresults, const char* argTypes, va_list args)
{
	assert(lua_isfunction(*m_state, m_stackIndex));
	LuaStackObject obj = m_state->GetGlobals_Stack()["_ALERT"];
	int startResults = lua_gettop(*m_state) + 1;
	Push();

	int numArgs = 0;
	const char* p = argTypes;
	while (*p != '\0')
	{
		switch (*p++)
		{
			case 'f':
				m_state->PushNumber(va_arg(args, float));
				break;
			case 'd':
				m_state->PushNumber(va_arg(args, double));
				break;
			case 'i':
				m_state->PushNumber(va_arg(args, int));
				break;
			case 's':
				m_state->PushString(va_arg(args, char*));
				break;
			case 'n':
				m_state->PushNil();
				break;
			case 'c':
				m_state->PushCClosure(va_arg(args, LuaStateCFunction), 0);
				break;
			case 'b':
				m_state->PushBoolean(va_arg(args, bool));
				break;
			case 'u':
				m_state->PushLightUserData(va_arg(args, void*));
				break;
			case 'o':
			{
				LuaStackObject obj = va_arg(args, LuaStackObject);
				obj.Push();
				break;
			}
			default:
				assert(0);
		}
		numArgs++;
	}

	int err = m_state->PCall(numArgs, nresults, obj);	err;
	assert(err == 0);
	obj.Pop();
	return LuaStackObject(m_state, startResults - 1);
}


LuaStackObject LuaStackObject::PCall(const char* argTypes, ...)
{
	va_list args;
	va_start(args, argTypes);

	LuaStackObject resultsObj = PCall(LUA_MULTRET, argTypes, args);

	va_end(args);

	return resultsObj;
}


LuaStackObject LuaStackObject::PCall(int nresults, const char* argTypes, ...)
{
	va_list args;
	va_start(args, argTypes);

	LuaStackObject resultsObj = PCall(nresults, argTypes, args);

	va_end(args);

	return resultsObj;
}

} // namespace LuaPlus

LUA_EXTERN_C_BEGIN
#include "ltable.h"
LUA_EXTERN_C_END

namespace LuaPlus {

LUA_API bool LuaPlusH_next(LuaState* state, LuaObject* table, LuaObject* key, LuaObject* value)
{
	Table* t = hvalue(&table->m_object);
	int i = luaH_index(*state, t, &key->m_object);  /* find original element */
	for (i++; i < t->sizearray; i++)
	{  /* try head array part */
		if (!ttisnil(&t->array[i]))
		{  /* a non-nil value? */
			key->AssignNumber(i + 1);
			value->AssignTObject(&t->array[i]);
			return true;
		}
	}
	for (i -= t->sizearray; i < sizenode(t); i++)
	{  /* then hash part */
		if (!ttisnil(val(node(t, i))))
		{  /* a non-nil value? */
			key->AssignTObject(key(node(t, i)));
			value->AssignTObject(val(node(t, i)));
			return true;
		}
	}
	return false;  /* no more elements */
}






int LuaState::Equal(const LuaObject& o1, const LuaObject& o2)
{
	int i = equalobj(o1.m_state->m_state, &o1.m_object, &o2.m_object);
	return i;
}


int LuaState::LessThan(const LuaObject& o1, const LuaObject& o2)
{
	int i = luaV_lessthan(o1.m_state->m_state, &o1.m_object, &o2.m_object);
	return i;
}


LuaObject LuaState::NewUserDataBox( void* u )
{
	LuaObject obj(this);
	obj.AssignUserData(u);
	return obj;
}


#define ismarked(x)	((x)->gch.marked & ((1<<4)|1))

#define markobject(st,o) { checkconsistency(o); \
  if (iscollectable(o) && !ismarked(gcvalue(o))) reallymarkobject(st,gcvalue(o)); }

void LuaPlusGCFunction(GCState* st)
{
	LuaState* state = (LuaState*)lua_getstateuserdata(st->L);

	LuaObject* curObj = state->GetHeadObject()->m_next;
	while (curObj != state->GetTailObject())
	{
		markobject(st, &curObj->m_object);
		curObj = curObj->m_next;		
	}
}




} // namespace LuaPlus
